import logging
import os

from airflow.operators.subdag_operator import SubDagOperator
from airflow.utils.log.file_task_handler import FileTaskHandler
from more_itertools import first_true

import dbnd

from dbnd import config, get_dbnd_project_config
from dbnd._core.constants import AD_HOC_DAG_PREFIX
from dbnd._core.context.databand_context import new_dbnd_context
from dbnd._core.log.buffered_log_manager import BufferedLogManager
from dbnd._core.tracking.airflow_dag_inplace_tracking import calc_task_key_from_af_ti
from dbnd._core.utils.basics import environ_utils
from dbnd._core.utils.uid_utils import get_task_run_attempt_uid_from_af_ti
from dbnd_airflow.tracking.execute_tracking import is_dag_eligable_for_tracking
from dbnd_airflow.tracking.fakes import FakeTaskRun


AIRFLOW_FILE_TASK_HANDLER = FileTaskHandler.__name__

AIRFLOW_TASK_LOGGER = "airflow.task"


class DbndAirflowHandler(logging.Handler):
    """
    This is a logger handler that is used as an Entry Point to airflow run.
    It's injected to the Logger(name="airflow.task"), and used by entering the context on the beginning of the task
    instance run, and getting close when the task instance is done.
    """

    def __init__(self, logger):
        logging.Handler.__init__(self)

        self.dbnd_context = None
        self.dbnd_context_manage = None

        self.task_run_attempt_uid = None
        self.task_env_key = None

        self.airflow_logger = logger

        self.in_memory_log_manager = None

    def set_context(self, ti):
        """
        Airflow's log handler use this method to setup the context when running a TaskInstance(=ti).
        We use this method to setup the dbnd context and communicate information to
        the `<airflow_operator>_execute` task, that we create in `execute_tracking.py`.
        """
        # we setting up only when we are not in our own orchestration dag
        if ti.dag_id.startswith(AD_HOC_DAG_PREFIX):
            return

        if not is_dag_eligable_for_tracking(ti.dag_id):
            return

        if config.getboolean("mlflow_tracking", "databand_tracking"):
            self.airflow_logger.warning(
                "dbnd can't track mlflow and airflow together please disable dbnd config "
                "`databand_tracking` in section `mlflow_tracking`"
            )
            return

        # we are not tracking SubDagOperator
        if ti.operator is None or ti.operator == SubDagOperator.__name__:
            return

        # Airflow is running with two process `run` and `--raw run`.
        # But we want the handler to run only once (Idempotency)
        # So we are using an environment variable to sync those two process
        task_key = calc_task_key_from_af_ti(ti)

        if ti.raw:
            self.in_memory_log_manager = BufferedLogManager()
            self.airflow_logger.debug(
                f"Initiated In Memory Log Manager with task {task_key}"
            )

        if os.environ.get(task_key, False):
            # This key is already set which means we are in `--raw run`
            return
        else:
            # We are in the outer `run`
            self.task_env_key = task_key
            # marking the environment with the current key for the
            environ_utils.set_on(task_key)
            from dbnd_airflow.tracking.dbnd_airflow_conf import (
                set_dbnd_config_from_airflow_connections,
            )

            # When we are in `--raw run`, in tracking, it runs the main airflow process
            # for every task, which made some of the features to run twice,
            # once when the `worker` process ran, and once when the `main` one ran,
            # which made some of the features to run with different configurations.
            # it still runs twice, but know with the same configurations.
            set_dbnd_config_from_airflow_connections()

        self.task_run_attempt_uid = get_task_run_attempt_uid_from_af_ti(ti)

        # make sure we are not polluting the airflow logs
        get_dbnd_project_config().quiet_mode = True

        # context with disabled logs
        self.dbnd_context_manage = new_dbnd_context(conf={"log": {"disabled": True}})
        self.dbnd_context = self.dbnd_context_manage.__enter__()

        # tracking msg
        self.airflow_logger.info(
            "Databand Tracking Started {version}".format(version=dbnd.__version__)
        )

    def close(self):
        if self.dbnd_context and self.in_memory_log_manager:
            try:
                fake_task_run = FakeTaskRun(
                    task_run_attempt_uid=self.task_run_attempt_uid
                )
                in_memory_log_body = self.in_memory_log_manager.get_log_body()
                self.dbnd_context.tracking_store.save_task_run_log(
                    task_run=fake_task_run, log_body=in_memory_log_body
                )
            except Exception:
                self.airflow_logger.exception("Exception occurred when saving task log")

        self.dbnd_context = None
        self.in_memory_log_manager = None

        try:
            if self.dbnd_context_manage:
                self.dbnd_context_manage.__exit__(None, None, None)
        except Exception:
            self.airflow_logger.exception(
                "Exception occurred when trying to exit dbnd_context_manager"
            )
        finally:
            self.dbnd_context_manage = None

        try:
            if self.task_env_key and self.task_env_key in os.environ:
                del os.environ[self.task_env_key]
        except Exception:
            self.airflow_logger.exception(
                "Exception occurred when trying to remove task_env_key from env"
            )
        finally:
            self.task_env_key = None

    def emit(self, record):
        if self.in_memory_log_manager:
            try:
                msg = self.format(record)
                self.in_memory_log_manager.add_log_msg(msg)
            except RecursionError:  # See issue 36272
                raise
            except Exception:
                self.handleError(record)


def set_dbnd_handler():
    """
    Build and inject the dbnd handler to airflow's logger.
    """
    airflow_logger = logging.getLogger(AIRFLOW_TASK_LOGGER)
    dbnd_handler = DbndAirflowHandler(logger=airflow_logger)

    base_file_handler = first_true(
        airflow_logger.handlers,
        pred=lambda handler: handler.__class__.__name__ == AIRFLOW_FILE_TASK_HANDLER,
        default=None,
    )
    if base_file_handler:
        handler_formatter = base_file_handler.formatter
        dbnd_handler.setFormatter(handler_formatter)
        for handler_filter in base_file_handler.filters:
            dbnd_handler.addFilter(handler_filter)
    airflow_logger.addHandler(dbnd_handler)



import logging

from collections import defaultdict
from typing import List, Union

import yaml

from airflow import DAG

from dbnd._core.configuration.dbnd_config import config
from dbnd._core.configuration.environ_config import ENV_DBND_DISABLE_SCHEDULED_DAGS_LOAD
from dbnd._core.utils.basics.environ_utils import environ_enabled
from dbnd._core.utils.string_utils import clean_job_name, pluralize
from dbnd._core.utils.timezone import convert_to_utc
from dbnd._vendor.marshmallow import Schema, fields
from dbnd.api.shared_schemas.scheduled_job_schema import validate_cron
from dbnd_airflow.scheduler.scheduler_dags_provider import DbndSchedulerOperator


logger = logging.getLogger(__name__)


class InvalidConfigException(Exception):
    pass


yaml_implicit_resolvers_no_timestamp = {
    k: [r for r in v if r[0] != "tag:yaml.org,2002:timestamp"]
    for k, v in yaml.SafeLoader.yaml_implicit_resolvers.items()
}


class JobFromFileSchema(Schema):
    class Meta:
        strict = True

    name = fields.Str(required=True)
    cmd = fields.Str(required=True)
    schedule_interval = fields.Str(required=True)
    start_date = fields.DateTime(allow_none=False, required=True, format="iso")
    owner = fields.Str(allow_none=False)

    end_date = fields.DateTime(allow_none=True, format="iso")
    depends_on_past = fields.Boolean(allow_none=True)

    catchup = fields.Boolean(allow_none=True)
    retries = fields.Int(allow_none=True)

    list_order = fields.Integer(allow_none=True)
    active = fields.Boolean(allow_none=True)


class DbndAirflowDagsProviderFromFile(object):
    def __init__(self, config_file, active_by_default=False, default_retries=1):
        config.load_system_configs()
        self.config_file = config_file
        self.active_by_default = active_by_default
        self.default_retries = default_retries

    def get_dags(self):  # type: () -> List[DAG]
        jobs = self.read_config()
        dags = []
        for job in jobs:
            dag = self.job_to_dag(job)
            dags.append(dag)
        return dags

    def read_config(self):
        from_file = self._read_config()

        if not from_file:
            from_file = []

        from_file = self.load_and_validate(from_file)

        for s in from_file:
            s["from_file"] = True
            s["active"] = s["active"] if "active" in s else self.active_by_default

        return from_file

    def _read_config(self):

        with open(self.config_file, "r") as stream:
            try:
                try:
                    # this horrible hack is because otherwise the yaml loader will decide to be "helpful" and convert anything that looks like a datetime
                    # to a datetime object. This later collides with marshmallow trying to parse and validate what it expects to be a datetime strings
                    NoDatesSafeLoader = yaml.SafeLoader
                    original_yaml_implicit_resolvers = (
                        NoDatesSafeLoader.yaml_implicit_resolvers
                    )
                    NoDatesSafeLoader.yaml_implicit_resolvers = (
                        yaml_implicit_resolvers_no_timestamp
                    )
                    config_entries = yaml.load(stream, Loader=NoDatesSafeLoader)
                finally:
                    NoDatesSafeLoader.yaml_implicit_resolvers = (
                        original_yaml_implicit_resolvers
                    )

                if not config_entries:
                    return []

                if not isinstance(config_entries, list):
                    raise InvalidConfigException(
                        "failed to load scheduler config at %s: invalid yaml format. Expecting a list of objects. Parsed type: %s"
                        % (self.config_file, config_entries.__class__.__name__)
                    )

                for i, s in enumerate(config_entries):
                    s["list_order"] = i
                return config_entries
            except yaml.YAMLError as exc:
                raise InvalidConfigException(
                    "failed to load scheduler config at %s: %s"
                    % (self.config_file, exc)
                )

    def load_and_validate(self, file_entries):
        schema = JobFromFileSchema(strict=False)

        # validate entries against the schema and check the schedule_interval
        loaded_entries = []
        for file_entry in file_entries:
            loaded_entry, errors = schema.load(file_entry)
            loaded_entries.append(loaded_entry)
            if errors:
                clean_errors = []
                for field, field_errors in errors.items():
                    clean_errors.append("%s: %s" % (field, ", ".join(field_errors)))
                loaded_entry["validation_errors"] = clean_errors
            else:
                loaded_entry["validation_errors"] = []

            if "schedule_interval" in loaded_entry:
                validation_error = validate_cron(loaded_entry["schedule_interval"])
                if validation_error:
                    loaded_entry["validation_errors"].append(
                        "Invalid schedule_interval: %s" % validation_error
                    )

        # check for duplicates
        group_by_name = defaultdict(lambda: [])
        for s in loaded_entries:
            if "name" in s:
                group_by_name[s["name"]].append(s)
        duplicates = [entries for entries in group_by_name.values() if len(entries) > 1]
        for duplicate_set in duplicates:
            for entry in duplicate_set:
                entry["validation_errors"].append(
                    "%s other %s exist in the configuration file with the same name"
                    % (
                        len(duplicate_set) - 1,
                        pluralize("entry", len(duplicate_set) - 1, "entries"),
                    )
                )

        # format and log validation errors
        errors = []
        for i, config_entry in enumerate(loaded_entries):
            if not config_entry["validation_errors"]:
                continue

            for validation_error in config_entry["validation_errors"]:
                msg = "JOB '%s': %s" % (
                    config_entry.get("name", "<entry #%s name missing>" % i),
                    validation_error,
                )
                errors.append(msg)

        if errors:
            raise InvalidConfigException(
                "scheduler config file %s validation errors:\n\t%s"
                % (self.config_file, "\n\t".join(errors))
            )
        return loaded_entries

    def job_to_dag(self, job):  # type: (dict) -> Union[DAG, None]

        default_args = {}
        if job.get("depends_on_past"):
            default_args["depends_on_past"] = job.get("depends_on_past")

        start_date = convert_to_utc(job.get("start_date"))
        if start_date:
            default_args["start_day"] = start_date

        if job.get("end_date"):
            default_args["end_date"] = convert_to_utc(job.get("end_date"))

        if job.get("owner"):
            default_args["owner"] = job.get("owner")

        job_name = clean_job_name(job["name"])
        dag = DAG(
            "%s" % job_name,
            start_date=start_date,
            default_args=default_args,
            schedule_interval=job.get("schedule_interval", None),
            catchup=job.get("catchup", False),
        )

        DbndSchedulerOperator(
            task_id="launcher",
            dag=dag,
            retries=job.get("retries") or self.default_retries,
            scheduled_cmd=job["cmd"],
            scheduled_job_name=job_name,
            with_name=False,
            scheduled_job_uid=job.get("uid", None),
            shell=config.getboolean("scheduler", "shell_cmd"),
        )

        return dag


def get_dags_from_file():
    if environ_enabled(ENV_DBND_DISABLE_SCHEDULED_DAGS_LOAD):
        return None

    try:
        # let be sure that we are loaded
        config.load_system_configs()

        config_file = config.get("scheduler", "config_file")
        if not config_file:
            logger.info("No dags file has been defined at scheduler.config_file")
            return {}
        default_retries = config.getint("scheduler", "default_retries")
        active_by_default = config.getboolean("scheduler", "active_by_default")

        dags = DbndAirflowDagsProviderFromFile(
            config_file=config_file,
            active_by_default=active_by_default,
            default_retries=default_retries,
        ).get_dags()
        return {dag.dag_id: dag for dag in dags}
    except Exception as e:
        logging.exception("Failed to get dags from the file")
        raise e



import logging

import flask_appbuilder

from airflow.plugins_manager import AirflowPlugin


logger = logging.getLogger(__name__)

flask_admin_views_list = []
flask_appbuilder_views_list = []

try:
    from airflow.models import DagModel
    from airflow.www.app import csrf as admin_csrf

    from dbnd_airflow.export_plugin.request_processing import (
        process_dag_run_states_data_request,
        process_full_runs_request,
        process_last_seen_values_request,
        process_metadata_request,
        process_new_runs_request,
    )
    from dbnd_airflow.export_plugin.utils import AIRFLOW_VERSION_2

    if not AIRFLOW_VERSION_2:
        import flask_admin

        from airflow.www.app import csrf as rbac_csrf
        from airflow.www_rbac.utils import CustomSQLAInterface

        class ExportDataViewAdmin(flask_admin.BaseView):
            def __init__(self, *args, **kwargs):
                super(ExportDataViewAdmin, self).__init__(*args, **kwargs)
                self.endpoint = "data_export_plugin"
                self.default_view = "new_runs"

            @admin_csrf.exempt
            @flask_admin.expose("/last_seen_values")
            def last_seen_values(self):
                return process_last_seen_values_request()

            @admin_csrf.exempt
            @flask_admin.expose("/")
            @flask_admin.expose("/new_runs", methods=["GET", "POST"])
            def new_runs(self):
                return process_new_runs_request()

            @admin_csrf.exempt
            @flask_admin.expose("/full_runs", methods=["GET", "POST"])
            def full_runs(self):
                return process_full_runs_request()

            @admin_csrf.exempt
            @flask_admin.expose("/runs_states_data", methods=["GET", "POST"])
            def task_instances(self):
                return process_dag_run_states_data_request()

            @admin_csrf.exempt
            @flask_admin.expose("/metadata")
            def metadata(self):
                return process_metadata_request()

        flask_admin_views_list.append(
            ExportDataViewAdmin(category="Admin", name="Export Data")
        )
        try:
            # this import is critical for loading `requires_authentication`
            from airflow import api

            api.load_auth()
            from airflow.www_rbac.api.experimental.endpoints import (
                api_experimental,
                csrf,
                requires_authentication,
            )

            @csrf.exempt
            @api_experimental.route("/last_seen_values", methods=["GET"])
            @requires_authentication
            def last_seen_values():
                return process_last_seen_values_request()

            @csrf.exempt
            @api_experimental.route("/new_runs", methods=["GET", "POST"])
            @requires_authentication
            def new_runs():
                return process_new_runs_request()

            @csrf.exempt
            @api_experimental.route("/full_runs", methods=["GET", "POST"])
            @requires_authentication
            def full_runs():
                return process_full_runs_request()

            @csrf.exempt
            @api_experimental.route("/runs_states_data", methods=["GET", "POST"])
            @requires_authentication
            def task_instances():
                return process_dag_run_states_data_request()

            @csrf.exempt
            @api_experimental.route("/metadata", methods=["GET"])
            @requires_authentication
            def metadata():
                return process_metadata_request()

        except Exception as e:
            logging.error("Export data could not be added to experimental api: %s", e)
    else:
        from airflow.www.app import csrf as rbac_csrf
        from airflow.www.utils import CustomSQLAInterface

    class ExportDataViewAppBuilder(flask_appbuilder.BaseView):
        endpoint = "data_export_plugin"
        default_view = "new_runs"
        # This line is required for the upgrade_check script as part of upgrading to Airflow 2.0
        # The actual model passed to CustomSQLAInterface doesn't really matter
        datamodel = CustomSQLAInterface(DagModel)

        @rbac_csrf.exempt
        @flask_appbuilder.has_access
        @flask_appbuilder.expose("/last_seen_values")
        def last_seen_values(self):
            return process_last_seen_values_request()

        @rbac_csrf.exempt
        @flask_appbuilder.has_access
        @flask_appbuilder.expose("/new_runs", methods=["GET", "POST"])
        def new_runs(self):
            return process_new_runs_request()

        @rbac_csrf.exempt
        @flask_appbuilder.has_access
        @flask_appbuilder.expose("/full_runs", methods=["GET", "POST"])
        def full_runs(self):
            return process_full_runs_request()

        @rbac_csrf.exempt
        @flask_appbuilder.has_access
        @flask_appbuilder.expose("/runs_states_data", methods=["GET", "POST"])
        def task_instances(self):
            return process_dag_run_states_data_request()

        @rbac_csrf.exempt
        @flask_appbuilder.has_access
        @flask_appbuilder.expose("/metadata")
        def metadata(self):
            return process_metadata_request()

    flask_appbuilder_views_list.append(
        {"category": "Admin", "name": "Export Data", "view": ExportDataViewAppBuilder()}
    )

except Exception:
    logger.error("Could not load export plugin", exc_info=True)


class DataExportAirflowPlugin(AirflowPlugin):
    name = "dbnd_airflow_export"
    admin_views = flask_admin_views_list
    appbuilder_views = flask_appbuilder_views_list



import logging
import typing

from collections import OrderedDict
from contextlib import contextmanager
from itertools import chain
from typing import Any, ContextManager, Dict, Optional

import six

from dbnd import dbnd_context
from dbnd._core.configuration.environ_config import ENV_DBND_SCRIPT_NAME
from dbnd._core.constants import TaskRunState, UpdateSource
from dbnd._core.tracking.airflow_dag_inplace_tracking import (
    get_task_run_uid_for_inline_script,
)
from dbnd._core.tracking.backends import TrackingStore
from dbnd._core.utils.type_check_utils import is_instance_by_class_name
from dbnd_airflow.tracking.conf_operations import flat_conf
from dbnd_airflow.tracking.dbnd_airflow_conf import get_databand_url_conf
from dbnd_airflow.tracking.dbnd_spark_conf import (
    add_spark_env_fields,
    dbnd_wrap_spark_environment,
    get_databricks_java_agent_conf,
    get_databricks_python_script_name,
    get_spark_submit_java_agent_conf,
    spark_submit_with_dbnd_tracking,
)
from dbnd_airflow.tracking.fakes import FakeRun, FakeTask, FakeTaskRun


if typing.TYPE_CHECKING:
    from airflow.contrib.hooks.databricks_hook import RunState as DatabricksRunState
    from airflow.contrib.operators.databricks_operator import (
        DatabricksHook,
        DatabricksSubmitRunOperator,
    )
    from airflow.contrib.operators.ecs_operator import ECSOperator
    from airflow.contrib.operators.spark_submit_operator import SparkSubmitOperator

logger = logging.getLogger(__name__)


@contextmanager
def track_emr_add_steps_operator(operator, tracking_info):
    logger.info("Tracking EmrAddStepsOperator")
    spark_envs = flat_conf(build_dbnd_spark_envs(tracking_info))
    for step in operator.steps:
        logger.info("Updating properties for %s", operator)
        args = step["HadoopJarStep"]["Args"]
        if args and "spark-submit" in args[0]:
            # Add dbnd and airflow context
            step["HadoopJarStep"]["Args"] = spark_submit_with_dbnd_tracking(
                args, dbnd_context=spark_envs
            )
            logger.info(
                "Properties for HadoopJarStep was updated for task %s", operator.task_id
            )
        else:
            logger.info("spark-submit has been not found in the operator")
    yield


def build_dbnd_spark_envs(tracking_info):
    # type: (Dict[str, str]) -> Dict[str, str]
    """
    Takes dbnd context as dict (airflow context + databand task context), builds spark.env properties,
    and enriches it with properties required for proper tracking in client and cluster modes
    :param tracking_info:
    :return:
    """
    return dict(
        chain(
            add_spark_env_fields(tracking_info).items(),
            dbnd_spark_conf().items(),
            spark_deploy_mode_cluster_conf().items(),
        )
    )


def dbnd_spark_conf():
    """
    Explicitly enable tracking
    :return:
    """
    return {"spark.env.DBND__TRACKING": "True"}


def spark_deploy_mode_cluster_conf():
    """
    Those variables should be passed when cluster is used for deploy mode
    :return:
    """
    return {
        "spark.yarn.appMasterEnv.DBND__TRACKING": "True",
        "spark.yarn.appMasterEnv.SPARK_ENV_LOADED": "1",
        "spark.yarn.appMasterEnv.DBND__ENABLE__SPARK_CONTEXT_ENV": "True",
        "spark.yarn.appMasterEnv.DBND_HOME": "/tmp/dbnd",
    }


@contextmanager
def track_databricks_submit_run_operator(operator, tracking_info):
    # type: (DatabricksSubmitRunOperator, Dict[str, str])-> None
    config = operator.json
    script_name = None  # type: str
    # passing env variables is only supported in new clusters
    if "new_cluster" in config:
        cluster = config["new_cluster"]
        cluster.setdefault("spark_env_vars", {})
        cluster["spark_env_vars"].update(tracking_info)
        cluster["spark_env_vars"].update(get_databand_url_conf())

        if "spark_python_task" in config:
            cluster["spark_env_vars"].update(
                get_databricks_python_script_name(
                    config["spark_python_task"]["python_file"]
                )
            )
            script_name = cluster["spark_env_vars"][ENV_DBND_SCRIPT_NAME]
            # calculate deterministic task uids so we can use it for manual completion
            (
                task_id,
                task_run_uid,
                task_run_attempt_uid,
            ) = get_task_run_uid_for_inline_script(tracking_info, script_name)

    if "spark_jar_task" in config:
        cluster.setdefault("spark_conf", {})
        agent_conf = get_databricks_java_agent_conf()
        if agent_conf is not None:
            cluster["spark_conf"].update(agent_conf)
    yield
    if script_name:
        # When dbnd is running inside Databricks, the script can be SIGTERM'd before dbnd will send state to tracker
        # So we need to check whenever the run was succeeded afterwards and set run state manually
        tracking_store = dbnd_context().tracking_store  # type: TrackingStore
        run = FakeRun(source=UpdateSource.airflow_tracking)
        task_run = FakeTaskRun(
            task_run_uid=task_run_uid,
            task_run_attempt_uid=task_run_attempt_uid,
            run=run,
            task=FakeTask(task_name=task_id, task_id=task_id),
            task_af_id=task_id,
        )
        try:
            hook = operator.get_hook()  # type: DatabricksHook
            state = hook.get_run_state(operator.run_id)  # type: DatabricksRunState
            run_page_url = hook.get_run_page_url(operator.run_id)
        except Exception as exc:
            logger.error(
                "Unable to get inline script run state from Databricks. Setting task run state to Failed: %s",
                exc,
            )
            set_task_run_state_safe(tracking_store, task_run, TaskRunState.FAILED)
            return

        save_external_links_safe(tracking_store, task_run, {"databricks": run_page_url})
        if state.is_successful:
            set_task_run_state_safe(tracking_store, task_run, TaskRunState.SUCCESS)
        else:
            # TODO: error should be extracted from plain Databricks logs
            set_task_run_state_safe(tracking_store, task_run, TaskRunState.FAILED)


def set_task_run_state_safe(tracking_store, task_run, state):
    # (TrackingStore, Any, TaskRunState) -> None
    try:
        tracking_store.set_task_run_state(task_run=task_run, state=state)
    except Exception as exc:
        logger.error("Unable to set task run state: %s", exc)


def save_external_links_safe(tracking_store, task_run, links_dict):
    # (TrackingStore, Any, Dict[str, str]) -> None
    try:
        tracking_store.save_external_links(
            task_run=task_run, external_links_dict=links_dict
        )
    except Exception as exc:
        logger.error("Unable to set external links: %s", exc)


@contextmanager
def track_dataproc_pyspark_operator(operator, tracking_info):
    if operator.dataproc_properties is None:
        operator.dataproc_properties = dict()
    spark_envs = build_dbnd_spark_envs(tracking_info)
    operator.dataproc_properties.update(spark_envs)
    yield


@contextmanager
def track_dataproc_submit_job_operator(operator, tracking_info):
    logger.info("Tracking DataprocSubmitJobOperator")
    if operator.job is not None:
        if "pyspark_job" in operator.job:
            logger.info("Updating properties for %s", operator)
            spark_envs = build_dbnd_spark_envs(tracking_info)

            pyspark_job = operator.job["pyspark_job"]
            pyspark_job.setdefault("properties", {})
            pyspark_job["properties"].update(spark_envs)
            logger.info(
                "Properties for pyspark_job was updated for task %s", operator.task_id
            )
        else:
            logger.info("pyspark_job has been not found in the operator")
    yield


@contextmanager
def track_spark_submit_operator(operator, tracking_info):
    # type: (SparkSubmitOperator, Dict[str,str])-> None

    if operator._conf is None:
        operator._conf = dict()
    spark_envs = build_dbnd_spark_envs(tracking_info)
    operator._conf.update(spark_envs)

    if operator._env_vars is None:
        operator._env_vars = dict()
    dbnd_env_vars = dbnd_wrap_spark_environment()
    operator._env_vars.update(dbnd_env_vars)

    if _has_java_application(operator):
        agent_conf = get_spark_submit_java_agent_conf()
        if agent_conf is not None:
            operator._conf.update(agent_conf)
    yield


def _has_java_application(operator):
    return (
        operator._application.endswith(".jar")
        or operator._jars
        and operator._jars.ends_with(".jar")
    )


@contextmanager
def track_ecs_operator(operator, tracking_info):
    # type: (ECSOperator, Dict[str,str])-> None
    """
    Adding the tracking info to the ECS environment through the `override` -> `containerOverrides`.
    Notice that we require to have `overrides` and `containerOverrides` with containers names in-ordered to make it work

    Airflow pass the override to boto so here is the boto3 docs:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecs.html#ECS.Client.run_task
    """
    info_as_env_var = [
        {"name": key, "value": value} for key, value in six.iteritems(tracking_info)
    ]

    new = []
    if "containerOverrides" in operator.overrides:
        for override in operator.overrides["containerOverrides"]:
            override.setdefault("environment", [])
            override["environment"].extend(info_as_env_var)
            new.append(override)

        operator.overrides["containerOverrides"] = new
    yield


# registering operators names to the relevant tracking method
_EXECUTE_TRACKING = OrderedDict(
    [
        ("EmrAddStepsOperator", track_emr_add_steps_operator),
        ("EmrPySparkOperator", track_emr_add_steps_operator),
        ("DatabricksSubmitRunOperator", track_databricks_submit_run_operator),
        # Airflow 1
        ("DataProcPySparkOperator", track_dataproc_pyspark_operator),
        # Airflow 2
        ("DataprocSubmitJobOperator", track_dataproc_submit_job_operator),
        ("DataprocSubmitPySparkJobOperator", track_dataproc_pyspark_operator),
        ("SparkSubmitOperator", track_spark_submit_operator),
        ("ECSOperator", track_ecs_operator),
    ]
)


def wrap_operator_with_tracking_info(tracking_info, operator):
    # type: (Dict[str, str], Any) -> Optional[ContextManager]
    """
    Wrap the operator with relevant tracking method, if found such method.
    """
    for class_name, tracking_wrapper in _EXECUTE_TRACKING.items():
        logger.debug(
            " %s %s %s",
            operator,
            class_name,
            is_instance_by_class_name(operator, class_name),
        )
        if is_instance_by_class_name(operator, class_name):
            return tracking_wrapper(operator, tracking_info)



from collections import defaultdict
from contextlib import contextmanager
from functools import wraps
from timeit import default_timer

import flask


class MetricCollector(object):
    def __init__(self):
        self.d = None

    @contextmanager
    def use_local(self):
        if self.d is not None:
            # already local mode
            yield self.d
            return

        self.d = defaultdict(dict)
        try:
            yield self.d
        finally:
            self.d = None

    def add(self, group, name, value):
        if self.d is not None:
            self.d[group][name] = value
        elif flask.has_app_context():
            metrics = getattr(flask.g, group, None)
            if metrics is None:
                metrics = {}
                setattr(flask.g, group, metrics)
            metrics[name] = value


METRIC_COLLECTOR = MetricCollector()


def measure_time(f):
    @wraps(f)
    def wrapped(*args, **kwargs):
        start = default_timer()
        result = f(*args, **kwargs)
        end = default_timer()
        METRIC_COLLECTOR.add("perf_metrics", f.__name__, end - start)
        return result

    return wrapped


def save_result_size(*names):
    def decorator(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            result = f(*args, **kwargs)
            values = result
            if len(names) == 1:
                values = (result,)
            for name, value_list in zip(names, values):
                METRIC_COLLECTOR.add("size_metrics", name, len(value_list))
            return result

        return wrapped

    return decorator



import logging

import six

from dbnd._core.configuration.environ_config import (
    DATABAND_AIRFLOW_CONN_ID,
    DBND_PARENT_TASK_RUN_ATTEMPT_UID,
    DBND_PARENT_TASK_RUN_UID,
    DBND_ROOT_RUN_TRACKER_URL,
    DBND_ROOT_RUN_UID,
    DBND_TRACE_ID,
)
from dbnd._core.settings import CoreConfig, TrackingConfig
from dbnd._core.utils.uid_utils import get_airflow_instance_uid
from dbnd.utils.trace import get_tracing_id
from dbnd_airflow.tracking.config import TrackingSparkConfig


AIRFLOW_DBND_CONNECTION_SOURCE = "airflow_dbnd_connection"

logger = logging.getLogger(__name__)


def get_airflow_conf(
    dag_id="{{dag.dag_id}}",
    task_id="{{task.task_id}}",
    execution_date="{{ts}}",
    try_number="{{task_instance._try_number}}",
):
    """
    These properties are
        AIRFLOW_CTX_DAG_ID - name of the Airflow DAG to associate a run with
        AIRFLOW_CTX_EXECUTION_DATE - execution_date to associate a run with
        AIRFLOW_CTX_TASK_ID - name of the Airflow Task to associate a run with
        AIRFLOW_CTX_TRY_NUMBER - try number of the Airflow Task to associate a run with
    """
    airflow_conf = {
        "AIRFLOW_CTX_DAG_ID": dag_id,
        "AIRFLOW_CTX_EXECUTION_DATE": execution_date,
        "AIRFLOW_CTX_TASK_ID": task_id,
        "AIRFLOW_CTX_TRY_NUMBER": try_number,
        "AIRFLOW_CTX_UID": get_airflow_instance_uid(),
    }
    airflow_conf.update(get_databand_url_conf())
    return airflow_conf


def _get_databand_url():
    try:
        external = TrackingConfig().databand_external_url
        if external:
            return external
        return CoreConfig().databand_url
    except Exception:
        pass


def get_databand_url_conf():
    databand_url = _get_databand_url()
    if databand_url:
        return {"DBND__CORE__DATABAND_URL": databand_url}
    return {}


def extract_airflow_tracking_conf(context):
    conf = extract_airflow_conf(context)
    conf.update(get_databand_url_conf())
    return conf


def extract_airflow_conf(context):
    task_instance = context.get("task_instance")
    if task_instance is None:
        return {}

    dag_id = task_instance.dag_id
    task_id = task_instance.task_id
    execution_date = task_instance.execution_date.isoformat()
    try_number = str(task_instance.try_number)

    if dag_id and task_id and execution_date:
        return {
            "AIRFLOW_CTX_DAG_ID": dag_id,
            "AIRFLOW_CTX_EXECUTION_DATE": execution_date,
            "AIRFLOW_CTX_TASK_ID": task_id,
            "AIRFLOW_CTX_TRY_NUMBER": try_number,
            "AIRFLOW_CTX_UID": get_airflow_instance_uid(),
        }
    return {}


def get_tracking_information(context, task_run):
    info = extract_airflow_conf(context)
    return extend_airflow_ctx_with_dbnd_tracking_info(task_run, info)


def extend_airflow_ctx_with_dbnd_tracking_info(task_run, airflow_ctx_env):
    info = airflow_ctx_env.copy()

    info[DBND_ROOT_RUN_UID] = task_run.run.root_run_info.root_run_uid
    info[DBND_ROOT_RUN_TRACKER_URL] = task_run.run.root_run_info.root_run_url
    info[DBND_PARENT_TASK_RUN_UID] = task_run.task_run_uid
    info[DBND_PARENT_TASK_RUN_ATTEMPT_UID] = task_run.task_run_attempt_uid
    info[DBND_TRACE_ID] = get_tracing_id().hex

    tracking_spark_conf = TrackingSparkConfig.from_databand_context()
    if tracking_spark_conf.provide_databand_service_endpoint:
        core = CoreConfig.from_databand_context()
        info["DBND__CORE__DATABAND_URL"] = core.databand_url
        info["DBND__CORE__DATABAND_ACCESS_TOKEN"] = core.databand_access_token

    info = {n: str(v) for n, v in six.iteritems(info) if v is not None}
    return info


def get_xcoms(task_instance):
    from airflow.models.xcom import XCom

    execution_date = task_instance.execution_date
    task_id = task_instance.task_id
    dag_id = task_instance.dag_id

    results = XCom.get_many(execution_date, task_ids=task_id, dag_ids=dag_id)
    return [(xcom.key, xcom.value) for xcom in results]


def get_dbnd_json_config_from_airflow_connections():
    """
    Set Databand config from Extra section in Airflow dbnd_config connection.
    Read about setting DBND Connection at: https://dbnd.readme.io/docs/setting-up-configurations-using-airflow-connections
    """
    from airflow.exceptions import AirflowException

    from dbnd_airflow.compat import BaseHook

    try:
        # Get connection from Airflow
        dbnd_conn_config = BaseHook.get_connection(DATABAND_AIRFLOW_CONN_ID)
        json_config = dbnd_conn_config.extra_dejson
        if json_config:
            return json_config

        if dbnd_conn_config.extra:
            # Airflow failed to parse extra config as json
            logger.error(
                "Extra config for {0} connection, should be formated as a valid json.".format(
                    DATABAND_AIRFLOW_CONN_ID
                )
            )

        else:
            # Extra section in connection is empty
            logger.warning(
                "No extra config provided to {0} connection.".format(
                    DATABAND_AIRFLOW_CONN_ID
                )
            )

        return None

    except AirflowException as afe:
        # Probably dbnd_config is not set properly in Airflow connections.
        logger.info(afe)
        return None
    except Exception:
        logger.exception("Failed to extract dbnd config from airflow's connection.")
        return None


def set_dbnd_config_from_airflow_connections():
    from dbnd._core.configuration.dbnd_config import config

    all_config_layers_names = set(
        [layer.name for layer in config.config_layer.get_all_layers()]
    )

    if AIRFLOW_DBND_CONNECTION_SOURCE not in all_config_layers_names:
        json_config = get_dbnd_json_config_from_airflow_connections()
        if not json_config:
            return False

        from dbnd._core.configuration.config_value import ConfigValuePriority

        config.set_values(
            config_values=json_config,
            priority=ConfigValuePriority.NORMAL,
            source=AIRFLOW_DBND_CONNECTION_SOURCE,
        )
        logger.debug(
            "Databand config was set using {0} connection.".format(
                DATABAND_AIRFLOW_CONN_ID
            )
        )
    return True



from __future__ import print_function

import logging
import os

from dbnd._core.utils.basics.format_exception import format_exception_as_str
from dbnd_airflow.utils import dbnd_airflow_path


logger = logging.getLogger(__name__)


def setup_scheduled_dags(sub_dir=None, unlink_first=True):
    from airflow import settings

    sub_dir = sub_dir or settings.DAGS_FOLDER

    if not sub_dir:
        raise Exception("Can't link scheduler: airflow's dag folder is undefined")

    source_path = dbnd_airflow_path("scheduler", "dags", "dbnd_dropin_scheduler.py")
    target_path = os.path.join(sub_dir, "dbnd_dropin_scheduler.py")

    if unlink_first and os.path.islink(target_path):
        try:
            logger.info("unlinking existing drop-in scheduler file at %s", target_path)
            os.unlink(target_path)
        except Exception:
            logger.error(
                "failed to unlink drop-in scheduler file at %s: %s",
                target_path,
                format_exception_as_str(),
            )
            return

    if not os.path.exists(target_path):
        try:
            logger.info("Linking %s to %s.", source_path, target_path)
            if not os.path.exists(sub_dir):
                os.makedirs(sub_dir)
            os.symlink(source_path, target_path)
        except Exception:
            logger.error(
                "failed to link drop-in scheduler in the airflow dags_folder: %s",
                format_exception_as_str(),
            )



from dbnd._core.log.dbnd_log import dbnd_log_exception, dbnd_log_init_msg
from dbnd_airflow.tracking.dbnd_dag_tracking import track_task


def _wrap_policy_with_dbnd_track_task(policy):
    if policy and getattr(policy, "_dbnd_patched", None):
        # already wrapped - don't double patch, return as is
        return policy

    def dbnd_track_task_policy(task):
        # Call original policy
        policy(task)
        # Wrap it with dbnd tracking
        track_task(task)

    dbnd_track_task_policy._dbnd_patched = True
    return dbnd_track_task_policy


def _patch_policy(module):
    if hasattr(module, "policy"):
        # airflow < 2.0, https://airflow.apache.org/docs/apache-airflow/1.10.10/concepts.html#cluster-policy
        new_policy = _wrap_policy_with_dbnd_track_task(module.policy)
        module.policy = new_policy
        dbnd_log_init_msg("patched module.policy with dbnd auto tracking")
    if hasattr(module, "task_policy"):
        # airflow >= 2.0, https://airflow.apache.org/docs/apache-airflow/stable/concepts/cluster-policies.html
        new_policy = _wrap_policy_with_dbnd_track_task(module.task_policy)
        module.task_policy = new_policy
        dbnd_log_init_msg("patched module.task_policy with dbnd auto tracking")


def _add_tracking_to_policy():
    try:
        # User can have this file or not
        import airflow_local_settings

        _patch_policy(airflow_local_settings)
        # we want to proceed and patch dagbag.settings as well
        # 1. _patch_policy is idempotent
        # 2. this code runs after local_settings are read
        # 3. there could be situation that local_settings exist but without policy
        #    (so global settings.policy is used anyway)
    except ImportError:
        pass

    from airflow.models.dagbag import settings

    _patch_policy(settings)


def add_tracking_to_policy():
    """Add tracking to all tasks as part of airflow policy"""
    try:
        _add_tracking_to_policy()
    except Exception as ex:
        dbnd_log_exception("Failed to add dbnd auto tracking to airflow policy %s" % ex)


def patch_airflow_context_vars():
    """Used for tracking bash operators"""

    from dbnd._core.utils.object_utils import patch_models
    from dbnd_airflow.airflow_override.operator_helpers import context_to_airflow_vars

    import airflow.models.taskinstance  # isort:skip

    modules_to_patch = [airflow.utils.operator_helpers, airflow.models.taskinstance]
    patches = []
    for module in modules_to_patch:
        if hasattr(module, "context_to_airflow_vars"):
            patches.append((module, "context_to_airflow_vars", context_to_airflow_vars))
    patch_models(patches)


def patch_snowflake_hook():
    # In order to use this patch the user need to have both `dbnd-snowflake` and `snowflake` installed
    try:
        import snowflake  # noqa: F401

        from airflow.contrib.hooks.snowflake_hook import SnowflakeHook

        from dbnd_snowflake.sql_tracking import (
            config_base_target_reporter,
            patch_airflow_db_hook,
        )
    except ImportError:
        # one of them is not available
        return

    patch_airflow_db_hook(SnowflakeHook, config_base_target_reporter)
